package WeddingGrid;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.Collections;
import java.util.Observable;
import java.util.Observer;

import repast.simphony.context.Context;
import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.random.RandomHelper;
import repast.simphony.space.continuous.ContinuousSpace;
import repast.simphony.space.continuous.NdPoint;
import repast.simphony.space.grid.Grid;
import repast.simphony.util.ContextUtils;

/**
 * A class that contain the initialiser and several common methods for different implementation of the world 
 * (eg. Billari's ring, the grid, ...)
 * @author James
 *
 */
public abstract class TheWorld implements Observer {

	protected final int startYear = 1950, maxCohortCount = 75;
	protected int[] totalHumanInAge, totalMarriedInAge, totalWomanInAge, totalMarriedWomanInAge;
	protected int[][] birthsByAgeAndYear, marriedBirthsByAgeAndYear, womenByAgeAndYear, marriedWomenByAgeAndYear, humanByAgeAndYear, marriagesByAgeAndYear;
	protected double[] hazardRate;
	protected ArrayList<Human> newborns;
	protected ArrayList<Cohort> cohorts;
	protected int cohortCount, year, idCount, ticksPerRun, birthCount, deathCount, marriageCount, marriageCountTotal = 0, illCount, agentTotal, maxAge, popsize, testMarriageCount, totalInitialMarriages,censusCheck, newill, unill, total16plus;
	protected StatisticCollector stc;
	protected boolean saveYearAgeStats;

	public TheWorld(Context<Object> context, ContinuousSpace<Object> space, Grid<Object> grid, StatisticCollector stc, boolean isYAxisAge) {
		this.stc = stc;
		//various count and state variables to initialise...
		cohortCount = 0;
		year = 1;
		idCount = 0;
		total16plus=0;
		birthCount = 0; deathCount = 0; marriageCount = 0; marriageCountTotal = 0; illCount=0; testMarriageCount=0; censusCheck=0; newill=0; unill=0;
		
		popsize = agentTotal = (Integer) RunEnvironment.getInstance().getParameters().getValue("initPop");
		ticksPerRun = (Integer) RunEnvironment.getInstance().getParameters().getValue("numOfTick") +1;
		
		newborns = new ArrayList<Human>();
		cohorts = new ArrayList<Cohort>();
		maxAge = (Integer) RunEnvironment.getInstance().getParameters().getValue("maxAge");
		if (maxAge == 0) maxAge = Data.defaultMaxAge;
		int seed = (int)  (Math.random()*100000);

		RandomHelper.setSeed(seed);
		
	
		hazardRate = new double[maxAge + 1];

		totalHumanInAge = new int[maxAge / 5 + 1];
		totalMarriedInAge = new int[maxAge / 5 + 1];
		totalWomanInAge = new int[maxAge / 5 + 1];
		totalMarriedWomanInAge = new int[maxAge / 5 + 1];
		
		// Arrays for detailed age/period outputs
		birthsByAgeAndYear = new int[maxAge + 1][ticksPerRun];
		marriedBirthsByAgeAndYear = new int[maxAge + 1][ticksPerRun];
		marriagesByAgeAndYear = new int[maxAge + 1][ticksPerRun];
		womenByAgeAndYear = new int[maxAge + 1][ticksPerRun];
		humanByAgeAndYear = new int[maxAge + 1][ticksPerRun];
		marriedWomenByAgeAndYear = new int[maxAge + 1][ticksPerRun];
		saveYearAgeStats = true;  // controls whether detailed year x age population and event csv records are created 
		
		
		int initialPopulation = (Integer) RunEnvironment.getInstance().getParameters().getValue("initPop");
		
		/*
		 * set up inital population based on empirical proportions in each age group
		 */

		for (int i = 0; i < initialPopulation; i++) {
			double r = RandomHelper.nextDoubleFromTo(0, 1);
			int a, b;
			if (r <= 0.29) { a = 0; b = 19; }
			else if (r <= 0.6) { a = 20; b = 39; }
			else if (r <= 0.84) { a = 40; b = 59; }
			else if (r <= 0.97) { a = 60; b = 79; }
			else { a = 80; b = 99; };

			int yob = RandomHelper.nextIntFromTo(startYear - b, startYear - a);
			
			r = RandomHelper.nextDoubleFromTo(0, 2.047);
			Human human;
			if (r < 1.048) human = new Male(context, space, grid, yob, startYear - yob, isYAxisAge);
			else {
				human = new Female(context, space, grid, yob, startYear - yob, isYAxisAge);
				totalWomanInAge[human.getAge() / 5]++;
			}
			totalHumanInAge[human.getAge() / 5]++;

			human.addObserver(this);
			human.setID(getID());

			if (isYAxisAge){
				double width = space.getDimensions().getWidth();
				space.moveTo(human.display, RandomHelper.nextDoubleFromTo(0, width), startYear - yob);
			}
			else {
				double width = space.getDimensions().getWidth();	
				double height = space.getDimensions().getHeight();
				space.moveTo(human.display, RandomHelper.nextDoubleFromTo(0, width), RandomHelper.nextDoubleFromTo(0, height));
			}
		}
		
		ArrayList<Human> marriedMale = new ArrayList<Human>(); 
		ArrayList<Human> marriedFemale = new ArrayList<Human>();
		
		/* 
		 * Determine whether setup population agents are married based on empirical proportions.
		 */
		for (Object obj : context) {
			if (obj instanceof Human) {
				double r = RandomHelper.nextDoubleFromTo(0, 1);
				if (r < Data.marriedRate((Human) obj)) {
					totalMarriedInAge[((Human) obj).getAge() / 5]++;
					marriageCountTotal++;
					if (obj instanceof Female) {
						totalMarriedWomanInAge[((Human) obj).getAge() / 5]++;
						marriedFemale.add((Human) obj);
					} else marriedMale.add((Human) obj);
				}
			}
			else {
				NdPoint pt = space.getLocation(obj);
				grid.moveTo(obj, (int) pt.getX(), (int) pt.getY());
			}
		}
		
		/*
		 * Match those agents who have been designated as married to each other,
		 * based on age.
		 * Assign a random marriage duration
		 */
		totalInitialMarriages=marriageCountTotal;
		Collections.sort(marriedMale, new AgeComparator());
		Collections.sort(marriedFemale, new AgeComparator());
		
		while ((marriedMale.size() > 0) && (marriedFemale.size() > 0)) {
			Human male = marriedMale.get(0);
			Human female = marriedFemale.get(0);
			
			int marDur = RandomHelper.nextIntFromTo(1, Math.min(male.age, female.age) - 16);
			male.setMarDur(marDur);
			male.partner = female;
			female.setMarDur(marDur);
			female.partner = male;
			if (!isYAxisAge) {
				male.moveToNewHome(1.0);
				female.moveToNewHome(1.0);
			}
			
			marriedMale.remove(male);
			marriedFemale.remove(female);		
		}
	}
	
	/**
	 * Perform various updates to the world when there is a change in the agent. 
	 * The type of update is determined by the type of the object the agent passed as 
	 * the argument when calling notifyObservers():
	 * <p>
	 * - null: normal yearly update.
	 * <p>
	 * - "Died": when an agent die.
	 * <p>
	 * - "Got married": when an agent get married.
	 * <p>
	 * - a Human object: when a new agent is born. The object is the newborn agent.
	 * 
	 * @param	human	the agent that called Observable.notifyObservers().
	 * @param	obj		the object passed as the argument.
	 * 
	 * @see 	Human, Observer, Observable.
	 * 
	 */
	public void update(Observable human, Object obj) {
		/* A birth - adjust various count variables accordingly */
		int a = ((Human) human).getAge();
		if (obj instanceof Human) {
			totalHumanInAge[0]++;
			birthCount++;
			agentTotal++;
			censusCheck++;
			popsize ++;
			
			birthsByAgeAndYear[a][year]++;
			if (((Human) human).isMarried()){
				marriedBirthsByAgeAndYear[a][year]++;
			}
			humanByAgeAndYear[0][year]++;
			if (obj instanceof Female) {
				totalWomanInAge[0]++;
				womenByAgeAndYear[0][year]++;
			}
			((Human) obj).addObserver(this);
			((Human) obj).setID(getID());

			if ((RunEnvironment.getInstance().isBatch() || (cohortCount <= maxCohortCount))) {
				newborns.add((Human)obj);
			}

		} else if (obj instanceof String) {
			
			if (((String) obj).equals("Got married")) {
				/* A marriage  - adjust various count variables accordingly*/
				marriageCount++;
				testMarriageCount++;
				totalMarriedInAge[a / 5]++;
				if (human instanceof Female) {
					marriagesByAgeAndYear[a][year]++;
					totalMarriedWomanInAge[a / 5]++;
				}
				
			} else if (((String) obj).equals("Died")) {
				/* A death  - adjust various count variables accordingly */
				deathCount++;
				popsize--;
				if (((Human) human).isIll) {illCount--; unill++;}
				if (((Human) human).isMarried()) {
					marriageCount--;
					totalMarriedInAge[a / 5]--;
					if (human instanceof Female) {
						totalMarriedWomanInAge[a / 5]--;
					}
				}
				totalHumanInAge[a / 5]--;
				if (human instanceof Female) {
					totalWomanInAge[a / 5]--;
				}
				stc.recordHuman((Human) human);
				
			} else if (((String) obj).equals("ill")) {
				 /* a health transition - adjust various count variables accordingly */
				illCount++;
				newill++;
			}
			

		} else {
			/* yearly update of every human  - adjust various count variables accordingly */
			censusCheck++;
			totalHumanInAge[(a - 1) / 5]--;
			totalHumanInAge[a / 5]++;
			humanByAgeAndYear[a][year]++;
			if (human instanceof Female) {
				totalWomanInAge[(a - 1) / 5]--;
				totalWomanInAge[a / 5]++;
				womenByAgeAndYear[a][year]++;
			}
			if (((Human) human).isMarried()) {
				totalMarriedInAge[(a - 1) / 5]--;
				totalMarriedInAge[a / 5]++;
				if (human instanceof Female) {
					totalMarriedWomanInAge[(a - 1) / 5]--;
					totalMarriedWomanInAge[a / 5]++;
					marriedWomenByAgeAndYear[a][year]++;
				}
			}
			if (human instanceof Female) {
				((Female) human).updateMarriedPercentage((double) totalMarriedWomanInAge[a / 5] / totalWomanInAge[a / 5]);
			}
			if (year==ticksPerRun-1){
				stc.recordHuman((Human) human);
			}
			if (((Human) human).age>15) {
				total16plus++;
			}
		}
	}
	
	
	/**
	 * Yearly update after all agents have been updated. Update hazard rate and record new data. 
	 * Create a new cohort if necessary.
	 * Poll agents for distance to nearest carer (if applicable)
	 * and check to see if agent is lonely
	 * 
	 * @see 	Cohort
	 */
	@ScheduledMethod(start = 1.5, interval = 1)
	public void eachYear() {
		if (year==61){
			/*create new carer distance array if year is 2011*/
			stc.carerDist= new ArrayList<Double>();
			/* print popsize  to screen*/
			System.out.println("popsize in 2011= " +  popsize);
		}
		//System.out.println("year = " + year);
		/* check lonely agents and distance to carer */
		Context<Object> context = ContextUtils.getContext(this);
		for (Object obj : context.getObjects(Human.class)) {
			stc.checkLonely((Human) obj);
			if (year==61){
				stc.checkCarer((Human) obj);
			}
		}
		
		
		/* check cohorts for completion and add new based on births */
		ArrayList<Cohort> completedCohorts = new ArrayList<Cohort>();
		for (Cohort cohort : cohorts) { 
			if (cohort.isFinished()) {
				completedCohorts.add(cohort);
			}
			else cohort.calculateHazardRate();
		}
		double[] newHazardRateData = new double[maxAge + 1];
		for (int i = 0; i < maxAge + 1; i++) newHazardRateData[i] = 0;
		for (Cohort cohort : completedCohorts) {
			double[] cHazardRate = cohort.getHazardRates();
			for (int i = 0; i < maxAge + 1; i++) newHazardRateData[i] += cHazardRate[i];
			cohorts.remove(cohort);
		}
		int completedCohortCount = cohortCount - cohorts.size();
		if ((completedCohorts.size() >= 0) && (++completedCohortCount >= 1)) {
			for (int i = 0; i < maxAge + 1; i++) 
				hazardRate[i] = (hazardRate[i] * (completedCohortCount - completedCohorts.size()) + newHazardRateData[i]) / completedCohortCount;
		}
		if ((RunEnvironment.getInstance().isBatch() || (cohortCount <= maxCohortCount))
				&& (newborns.size() > 0)) {
			cohorts.add(new Cohort(newborns));
			newborns = new ArrayList<Human>();
			cohortCount++;
		}
		
		/* record hazard rates and peak hazard value */ 
		if ( (RunEnvironment.getInstance().isBatch() && (year+1 == (Integer) RunEnvironment.getInstance().getParameters().getValue("numOfTick")))) {
			stc.updateHazardRate(hazardRate);
			double maxHaz=0; int maxIndex=0;
			for (int i=0; i<maxAge; i++ ){
				if (hazardRate[i]>maxHaz){
					maxHaz=hazardRate[i]; maxIndex=i;
				}
			}
			stc.hazardPeak=maxIndex;
		} 
		
		
		/* yearly recording stats; set up if year =1 , else record */ 
		if ((year == 1))
			stc.newYearlyCountRecord();
		marriageCountTotal += marriageCount;
		if ((year <=ticksPerRun-1))
			stc.recordYearlyCounts(marriageCountTotal, illCount, agentTotal, popsize,total16plus);
		
		
		stc.recordPerYear(startYear + year, birthCount, deathCount);
		
		/* record all other stats if we have reached the end of the sim */ 
		if ((year == ticksPerRun-1)) {
			//stc.newMucmRecord();
			//stc.recordMucm(marriageCountTotal, agentTotal,testMarriageCount);
			if (saveYearAgeStats){
				stc.recordBirthStats(birthsByAgeAndYear);
				stc.recordMarriedBirthStats(marriedBirthsByAgeAndYear);
				stc.recordMarriageStats(marriagesByAgeAndYear);
				stc.recordPopStats(humanByAgeAndYear);
				stc.recordWomenStats(womenByAgeAndYear);
				stc.recordMarriedStats(marriedWomenByAgeAndYear);
				stc.recordCarerStats(stc.carerDist);
			}
			
			System.out.println(" run " + StatisticCollector.runCount );
			WeddingGrid.StatisticCollector.runCount++;
			/*if (StatisticCollector.runCount % <some integer reflecting the number of parameter combinations> == 0) {
				StatisticCollector.set++;
			}*/
		}
		
	

		
		/* uncomment for decade hazard reporting 
		*if ((year % 10 == 0)) stc.updateDecadeHazardRate(hazardRate);
		*/
		
		/* increment year and reset counts */
		year++;
		
		birthCount = 0; deathCount = 0; marriageCount = 0; birthCount = 0; deathCount = 0; newill=0; unill=0; total16plus=0; 
		WeddingGrid.StatisticCollector.lonelyCount=0;  
		WeddingGrid.StatisticCollector.lonelyWidowed=0;
		WeddingGrid.StatisticCollector.lonelyKidsDead=0;
		WeddingGrid.StatisticCollector.caredByKids=0;
		WeddingGrid.StatisticCollector.caredByPartner=0;
		
	}

	/**
	 * Return the hazard rates.
	 * 
	 * @return		An array contain the hazard rates with respect to ages.
	 */
	public double[] getHazardRates() {
		return hazardRate;
	}

	/**
	 * Return the current year.
	 * 
	 * @return		The current year.
	 */
	public int getYear() {
		return year;
	}

	/**
	 * Get an ID for the new agent.
	 *  
	 * @return 		ID for the new agent.
	 */
	public synchronized int getID() {
		return ++idCount;
	}
}

class AgeComparator implements Comparator<Human> {
	
	@Override
	public int compare(Human h1, Human h2) {
		return (h1.age > h2.age ? -1 : (h1.age == h2.age ? 0 : 1));
	}
}
